#include "sqlitebackend.h"



/*
 *  Declaration of SQLite capabilities
 */

static struct odbx_basic_ops sqlite_odbx_basic_ops = {
	.init = sqlite_odbx_init,
	.bind_simple = sqlite_odbx_bind_simple,
	.unbind = sqlite_odbx_unbind,
	.finish = sqlite_odbx_finish,
	.get_option = sqlite_odbx_get_option,
	.set_option = sqlite_odbx_set_option,
	.error = sqlite_odbx_error,
	.escape = sqlite_odbx_escape,
	.query = sqlite_odbx_query,
	.result = sqlite_odbx_result,
	.result_free = sqlite_odbx_result_free,
	.rows_affected = sqlite_odbx_rows_affected,
	.row_fetch = sqlite_odbx_row_fetch,
	.column_count = sqlite_odbx_column_count,
	.column_name = sqlite_odbx_column_name,
	.column_type = sqlite_odbx_column_type,
	.field_length = sqlite_odbx_field_length,
	.field_value = sqlite_odbx_field_value,
};



static struct odbx_ops sqlite_odbx_ops = {
	.basic = &sqlite_odbx_basic_ops
};



void odbxdrv_register( struct odbx_ops** ops )
{
	*ops = &sqlite_odbx_ops;
}



/*
 *  Private sqlite error messages
 */

static const char* sqlite_odbx_errmsg[] = {
	gettext_noop("Opening database failed")
};



/*
 *  ODBX basic operations
 *  SQLite style
 */


/*
 *  SQLite doesn't connect to a database server. Instead it opens the database
 *  (a file on the harddisk) directly. Thus, host doesn't contain the name of a
 *  computer but of a directory in the filesystem instead. Port is unused.
 */

static int sqlite_odbx_init( odbx_t* handle, const char* host, const char* port )
{
	if( ( handle->aux = malloc( sizeof( struct sconn ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	struct sconn* aux = handle->aux;

	aux->pathlen = 0;
	aux->errmsg = NULL;
	handle->generic = NULL;

	if( host != NULL )
	{
		aux->pathlen = strlen( host );   /* host == directory */

		if( aux->pathlen > SQLITE_BUFLEN )
		{
			aux->pathlen = 0;
			free( handle->aux );
			handle->aux = NULL;

			return -ODBX_ERR_TOOLONG;
		}

		memcpy( aux->path, host, aux->pathlen + 1 );
	}

	return ODBX_ERR_SUCCESS;
}



/*
 *  username and password are not used by SQLite. It relies on the filesystem
 *  access control mechanisms for allowing or preventing database access.
 */

static int sqlite_odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password )
{
	int err;


	if( !(handle->aux && database) )
	{
		return -ODBX_ERR_PARAM;
	}

	size_t flen = strlen( database ) + 1;
	struct sconn* aux = handle->aux;

	aux->errmsg = NULL;

	if( ( aux->pathlen + flen ) > SQLITE_BUFLEN )
	{
		return -ODBX_ERR_TOOLONG;
	}

	memcpy( aux->path + aux->pathlen, database, flen );

	/*  The second parameter is currently unused. */
	if( ( handle->generic = (void*) sqlite_open( aux->path, 0, NULL ) ) == NULL )
	{
		aux->errmsg = (char*) dgettext( "opendbx", sqlite_odbx_errmsg[0] );
		return -ODBX_ERR_BACKEND;
	}

	if( ( err = sqlite_exec( (sqlite*) handle->generic, "PRAGMA empty_result_callbacks = ON;",
		NULL, NULL, NULL ) ) != SQLITE_OK )
	{
		aux->errmsg = (char*) sqlite_error_string( err );
		return -ODBX_ERR_BACKEND;
	}

	return ODBX_ERR_SUCCESS;
}



static int sqlite_odbx_unbind( odbx_t* handle )
{
	if( !(handle->generic && handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	((struct sconn*) handle->aux)->errmsg = NULL;

	sqlite_close( (sqlite*) handle->generic );
	handle->generic = NULL;

	return ODBX_ERR_SUCCESS;
}



static int sqlite_odbx_finish( odbx_t* handle )
{
	if( !(handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	free(  handle->aux );
	handle->aux = NULL;

	return ODBX_ERR_SUCCESS;
}



static int sqlite_odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	if( !(handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	((struct sconn*) handle->aux)->errmsg = NULL;

	switch( option )
	{
		case ODBX_OPT_API_VERSION:
			*(int*) value = 0;
			break;
		case ODBX_OPT_THREAD_SAFE:
			*(int*) value = 0;   /* FIXME: How to find out if THREADSAFE was set while sqlite compilation */
			break;
		case ODBX_OPT_TLS:
			*(int*) value = 0;
			break;
		case ODBX_OPT_MULTI_STATEMENTS:
			*(int*) value = 0;
			break;
		case ODBX_OPT_PAGED_RESULTS:
			*(int*) value = 0;
			break;
		case ODBX_OPT_COMPRESS:
			*(int*) value = 0;
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static int sqlite_odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	if( !(handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	((struct sconn*) handle->aux)->errmsg = NULL;

	switch( option )
	{
		case ODBX_OPT_API_VERSION:
		case ODBX_OPT_THREAD_SAFE:
			return -ODBX_ERR_OPTRO;
		case ODBX_OPT_TLS:
		case ODBX_OPT_MULTI_STATEMENTS:
		case ODBX_OPT_PAGED_RESULTS:
		case ODBX_OPT_COMPRESS:
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return -ODBX_ERR_OPTWR;
}



static const char* sqlite_odbx_error( odbx_t* handle )
{
	if( handle->aux )
	{
		return ((struct sconn*) handle->aux)->errmsg;
	}

	return NULL;
}



/*
 *  SQLite doesn't provide an escape function for values in the SQL statement.
 *  Therefore it is implemented by the driver itself.
 */

static int sqlite_odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	unsigned long i, len = 0;


	if( !(handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	((struct sconn*) handle->aux)->errmsg = NULL;

	for( i = 0; i < fromlen; i++ )
	{
		if( i < *tolen - 2 )
		{
			switch( from[i] )
			{
				case '\'':   /* SQL doubling */
					*to++ = '\'';
					len++;
					break;
			}
			*to++ = from[i];
			len++;
		}
		else
		{
			return -ODBX_ERR_TOOLONG;
		}
	}

	to[len] = '\0';
	*tolen = len;

	return ODBX_ERR_SUCCESS;
}



static int sqlite_odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	if( !(handle->aux && query) )
	{
		return -ODBX_ERR_PARAM;
	}

	((struct sconn*) handle->aux)->errmsg = NULL;

	if( ( ((struct sconn*) handle->aux)->stmt = malloc( length + 1 ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	memcpy( ((struct sconn*) handle->aux)->stmt, query, length );
	((struct sconn*) handle->aux)->stmt[length] = '\0';

	return ODBX_ERR_SUCCESS;
}



static int sqlite_odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	char** res;
	long ms = 0;
	int err, nrow, ncolumn;
	struct sres* sres;
	struct sconn* aux = (struct sconn*) handle->aux;


	if( !(handle->generic && handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	aux->errmsg = NULL;
	if( aux->stmt == NULL )
	{
		return 0;   /* no more results */
	}

	if( timeout != NULL )
	{
		ms = timeout->tv_sec * 1000 + timeout->tv_usec;
	}

	while( ( err = sqlite_get_table( (sqlite*) handle->generic, aux->stmt, &res, &nrow, &ncolumn, NULL ) ) == SQLITE_BUSY )
	{
		if( ms <= 0 ) {
			return 1;   /* Timeout */
		}
		sqlite_busy_timeout( (sqlite*) handle->generic, 100 );
		ms -= 100;
	}

	free( aux->stmt );
	aux->stmt = NULL;

	if( err != SQLITE_OK )
	{
		aux->errmsg = (char*) sqlite_error_string( err );
		return -ODBX_ERR_BACKEND;
	}

	if( ( *result = (odbx_result_t*) malloc( sizeof( struct odbx_result_t ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	if( ( sres = (struct sres*) malloc( sizeof( struct sres ) ) ) == NULL )
	{
		free( *result );
		*result = NULL;

		return -ODBX_ERR_NOMEM;
	}

	(*result)->generic = (void*) res;
	(*result)->aux = (void*) sres;
	sres->ncolumn = ncolumn;
	sres->nrow = nrow;
	sres->cur = -1;

	if( ncolumn == 0 )
	{
		return 2;   /* empty or not SELECT like query */
	}

	return 3;   /* result is available */
}



static void sqlite_odbx_result_free( odbx_result_t* result )
{
	if( result->handle && result->handle->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
	}

	sqlite_free_table( (char**) result->generic );

	if( result->aux )
	{
		free( result->aux );
		result->aux = NULL;
	}

	free( result );
}



static int sqlite_odbx_row_fetch( odbx_result_t* result )
{
	struct sres* res = (struct sres*) result->aux;

	if( !(result->aux && result->handle && result->handle->aux) )
	{
		return -ODBX_ERR_PARAM;
	}

	((struct sconn*) result->handle->aux)->errmsg = NULL;

	res->cur++;
	if( res->cur < res->nrow ) { return 1; }

	return 0;
}



static uint64_t sqlite_odbx_rows_affected( odbx_result_t* result )
{
	if( result->handle && result->handle->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
		return (uint64_t) sqlite_changes( (sqlite*) result->handle->generic );
	}

	return 0;
}



static unsigned long sqlite_odbx_column_count( odbx_result_t* result )
{
	if( result->handle && result->handle->aux && result->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
		return ((struct sres*) result->aux)->ncolumn;
	}

	return 0;
}



static const char* sqlite_odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	if( result->handle && result->handle->aux && result->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
		if( result->generic && pos < ((struct sres*) result->aux)->ncolumn )
		{
			return ((const char**) result->generic)[pos];
		}
	}

	return NULL;
}



static int sqlite_odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	if( result->handle && result->handle->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
		return ODBX_CLOB;
	}

	return -ODBX_ERR_PARAM;
}



static unsigned long sqlite_odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	int num;

	if( result->handle && result->handle->aux && result->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
		if( result->generic && pos < ((struct sres*) result->aux)->ncolumn )
		{
			num = ((struct sres*) result->aux)->cur * ((struct sres*) result->aux)->ncolumn + ((struct sres*) result->aux)->ncolumn + pos;
			return (unsigned long) strlen( ((char**) result->generic)[num] );
		}
	}

	return 0;
}



static const char* sqlite_odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	int num;

	if( result->handle && result->handle->aux && result->aux )
	{
		((struct sconn*) result->handle->aux)->errmsg = NULL;
		if( result->generic && pos < ((struct sres*) result->aux)->ncolumn )
		{
			num = ((struct sres*) result->aux)->ncolumn + ((struct sres*) result->aux)->cur * ((struct sres*) result->aux)->ncolumn + pos;
			return  ((const char**) result->generic)[num];
		}
	}

	return NULL;
}
