#include "sqlite3backend.h"



/*
 *  Declaration of SQLite capabilities
 */

static struct odbx_basic_ops sqlite3_odbx_basic_ops = {
	.init = sqlite3_odbx_init,
	.bind_simple = sqlite3_odbx_bind_simple,
	.unbind = sqlite3_odbx_unbind,
	.finish = sqlite3_odbx_finish,
	.get_option = sqlite3_odbx_get_option,
	.set_option = sqlite3_odbx_set_option,
	.error = sqlite3_odbx_error,
	.escape = sqlite3_odbx_escape,
	.query = sqlite3_odbx_query,
	.result = sqlite3_odbx_result,
	.result_free = sqlite3_odbx_result_free,
	.rows_affected = sqlite3_odbx_rows_affected,
	.row_fetch = sqlite3_odbx_row_fetch,
	.column_count = sqlite3_odbx_column_count,
	.column_name = sqlite3_odbx_column_name,
	.column_type = sqlite3_odbx_column_type,
	.field_length = sqlite3_odbx_field_length,
	.field_value = sqlite3_odbx_field_value,
};



static struct odbx_ops sqlite3_odbx_ops = {
	.basic = &sqlite3_odbx_basic_ops,
};



void odbxdrv_register( struct odbx_ops** ops )
{
	*ops = &sqlite3_odbx_ops;
}



/*
 *  ODBX basic operations
 *  SQLite3 style
 */


/*
 *  SQLite3 doesn't connect to a database server. Instead it opens the database
 *  (a file on the harddisk) directly. Thus, host doesn't contain the name of a
 *  computer but of a directory in the filesystem instead. Port is unused.
 */

static int sqlite3_odbx_init( odbx_t* handle, const char* host, const char* port )
{
	if( ( handle->aux = malloc( sizeof( struct sconn ) ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	struct sconn* conn = (struct sconn*) handle->aux;

	conn->pathlen = 0;
	handle->generic = NULL;

	if( host != NULL )
	{
		conn->pathlen = strlen( host );   /* host == directory */

		if( conn->pathlen > SQLITE3_BUFLEN )
		{
			conn->pathlen = 0;
			free( handle->aux );
			handle->aux = NULL;

			return -ODBX_ERR_TOOLONG;
		}

		memcpy( conn->path, host, conn->pathlen + 1 );
	}

	return ODBX_ERR_SUCCESS;
}



/*
 *  username and password are not used by SQLite. It relies on the filesystem
 *  access control mechanisms for allowing or preventing database access.
 */

static int sqlite3_odbx_bind_simple( odbx_t* handle, const char* database, const char* username, const char* password )
{
	size_t len;
	struct sconn* conn;


	if( !(handle->aux && database) )
	{
		return -ODBX_ERR_PARAM;
	}

	conn = (struct sconn*) handle->aux;
	len = conn->pathlen;

	len += snprintf( conn->path + len, SQLITE3_BUFLEN - len, "%s", database );

	if( len > SQLITE3_BUFLEN )
	{
		return -ODBX_ERR_TOOLONG;
	}
	conn->path[len] = '\0';

	sqlite3* s3conn;
	if(  sqlite3_open( conn->path, &s3conn ) != SQLITE_OK )
	{
		return -ODBX_ERR_BACKEND;
	}
	handle->generic = (void*) s3conn;

	return ODBX_ERR_SUCCESS;
}



static int sqlite3_odbx_unbind( odbx_t* handle )
{
	if( sqlite3_close( (sqlite3*) handle->generic ) != SQLITE_OK )
	{
		return -ODBX_ERR_BACKEND;
	}

	handle->generic = NULL;
	return ODBX_ERR_SUCCESS;
}



static int sqlite3_odbx_finish( odbx_t* handle )
{
	if( handle->aux )
	{
		free(  handle->aux );
		handle->aux = NULL;

		return ODBX_ERR_SUCCESS;
	}

	return -ODBX_ERR_PARAM;
}



static int sqlite3_odbx_get_option( odbx_t* handle, unsigned int option, void* value )
{
	switch( option )
	{
		case ODBX_OPT_API_VERSION:
			*(int*) value = 0;
			break;
		case ODBX_OPT_THREAD_SAFE:
			*(int*) value = 0;   /* FIXME: How to find out if THREADSAFE was set while sqlite3 compilation */
			break;
		case ODBX_OPT_TLS:
			*(int*) value = 0;
			break;
		case ODBX_OPT_MULTI_STATEMENTS:
			*(int*) value = 1;
			break;
		case ODBX_OPT_PAGED_RESULTS:
			*(int*) value = 0;
			break;
		case ODBX_OPT_COMPRESS:
			*(int*) value = 0;
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return ODBX_ERR_SUCCESS;
}



static int sqlite3_odbx_set_option( odbx_t* handle, unsigned int option, void* value )
{
	switch( option )
	{
		case ODBX_OPT_API_VERSION:
		case ODBX_OPT_THREAD_SAFE:
			return -ODBX_ERR_OPTRO;
		case ODBX_OPT_MULTI_STATEMENTS:
			if( *((int*) value) == 1 )
			{
				return ODBX_ERR_SUCCESS;
			}
			break;
		case ODBX_OPT_PAGED_RESULTS:
		case ODBX_OPT_COMPRESS:
		case ODBX_OPT_TLS:
			break;
		default:
			return -ODBX_ERR_OPTION;
	}

	return -ODBX_ERR_OPTWR;
}



static const char* sqlite3_odbx_error( odbx_t* handle )
{
	return sqlite3_errmsg( (sqlite3*) handle->generic );
}



/*
 *  SQLite3 doesn't provide an escape function for values in the SQL statement.
 *  Therefore it is implemented by the driver itself.
 */

static int sqlite3_odbx_escape( odbx_t* handle, const char* from, unsigned long fromlen, char* to, unsigned long* tolen )
{
	unsigned long i, len = 0;


	for( i = 0; i < fromlen; i++ )
	{
		if( i < *tolen - 2 )
		{
			switch( from[i] )
			{
				case '\'':   /* SQL doubling */
					*to++ = '\'';
					len++;
					break;
			}
			*to++ = from[i];
			len++;
		}
		else
		{
			return -ODBX_ERR_TOOLONG;
		}
	}

	to[len] = '\0';
	*tolen = len;

	return ODBX_ERR_SUCCESS;
}



static int sqlite3_odbx_query( odbx_t* handle, const char* query, unsigned long length )
{
	struct sconn* sc = (struct sconn*) handle->aux;

	if( !(query && sc) )
	{
		return -ODBX_ERR_PARAM;
	}

	if( ( sc->stmt = malloc( length + 1 ) ) == NULL )
	{
		return -ODBX_ERR_NOMEM;
	}

	memcpy( sc->stmt, query, length );

	sc->stmt[length] = '\0';
	sc->length = length;
	sc->tail = sc->stmt;

	return ODBX_ERR_SUCCESS;
}



static int sqlite3_odbx_result( odbx_t* handle, odbx_result_t** result, struct timeval* timeout, unsigned long chunk )
{
	sqlite3_stmt* res;
	struct sconn* aux = (struct sconn*) handle->aux;


	if( aux == NULL )
	{
		return -ODBX_ERR_PARAM;
	}

	if( aux->stmt == NULL )
	{
		return 0;   /* no more results */
	}

	if( aux->tail[0] == '\0' )
	{
		free( aux->stmt );
		aux->stmt = NULL;

		return 0;   /* no more results */
	}

	if( ( *result = (odbx_result_t*) malloc( sizeof( struct odbx_result_t ) ) ) == NULL )
	{
		free( aux->stmt );
		aux->stmt = NULL;

		return -ODBX_ERR_NOMEM;
	}

	if( timeout != NULL )
	{
		sqlite3_busy_timeout( handle->generic, timeout->tv_sec * 1000 + timeout->tv_usec );
	}

	// FIXME: Using timeout this way is not 100% clean

	if( sqlite3_prepare( (sqlite3*) handle->generic, aux->tail, -1, &res, (const char**) &(aux->tail) ) != SQLITE_OK )
	{
		free( aux->stmt );
		aux->stmt = NULL;

		free( *result );
		*result = NULL;

		return -ODBX_ERR_BACKEND;
	}

	(*result)->generic = res;

	if( sqlite3_column_count( res ) == 0 )
	{
		sqlite3_step( res );
		return 2;   /* empty or not SELECT like query */
	}

	return 3;   /* result is available */
}



static void sqlite3_odbx_result_free( odbx_result_t* result )
{
	if( result->generic )
	{
		sqlite3_finalize( (sqlite3_stmt*) result->generic );
		result->generic = NULL;
	}

	free( result );
}



static int sqlite3_odbx_row_fetch( odbx_result_t* result )
{
	switch( sqlite3_step( (sqlite3_stmt*) result->generic ) )
	{
		case SQLITE_ROW:
			return 1;
		case SQLITE_DONE:
			return 0;
	}

	return -ODBX_ERR_BACKEND;
}



static uint64_t sqlite3_odbx_rows_affected( odbx_result_t* result )
{
	if( result->handle )
	{
		return (uint64_t) sqlite3_changes( (sqlite3*) result->handle->generic );
	}

	return 0;
}



static unsigned long sqlite3_odbx_column_count( odbx_result_t* result )
{
	if( result->generic )
	{
		return (unsigned long) sqlite3_column_count( (sqlite3_stmt*) result->generic );
	}

	return 0;
}



static const char* sqlite3_odbx_column_name( odbx_result_t* result, unsigned long pos )
{
	if( result->generic && pos < sqlite3_column_count( (sqlite3_stmt*) result->generic ) )
	{
		return (const char*) sqlite3_column_name( (sqlite3_stmt*) result->generic, pos );
	}

	return NULL;
}



static int sqlite3_odbx_column_type( odbx_result_t* result, unsigned long pos )
{
	if( result->generic && pos < sqlite3_column_count( (sqlite3_stmt*) result->generic ) )
	{
		switch( sqlite3_column_type( (sqlite3_stmt*) result->generic, pos ) )
		{
			case SQLITE_INTEGER:
				return ODBX_BIGINT;
			case SQLITE_FLOAT:
				return ODBX_DOUBLE;
			case SQLITE_BLOB:
				return ODBX_BLOB;
			case SQLITE_TEXT:
				return ODBX_CLOB;
		}

		return ODBX_UNKNOWN;
	}

	return -ODBX_ERR_PARAM;
}



static unsigned long sqlite3_odbx_field_length( odbx_result_t* result, unsigned long pos )
{
	if( result->generic && pos < sqlite3_column_count( (sqlite3_stmt*) result->generic ) )
	{
		return (unsigned long) sqlite3_column_bytes( (sqlite3_stmt*) result->generic, pos );
	}

	return 0;
}



static const char* sqlite3_odbx_field_value( odbx_result_t* result, unsigned long pos )
{
	if( result->generic && pos < sqlite3_column_count( (sqlite3_stmt*) result->generic ) )
	{
		return (const char*) sqlite3_column_blob( (sqlite3_stmt*) result->generic, pos );
	}

	return NULL;
}
