#include <inttypes.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <odbx.h>



static const char* mysql_basic[] = {
	"CREATE TABLE odbxtest ( i8 tinyint, i16 smallint, i32 int, i64 bigint, f4 float, f8 double, string varchar(25) );",
	"INSERT INTO odbxtest VALUES ( 100, 1000, 100000, 10000000, 10000.1, 10000000.1, 'first string' );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, 'varstring' );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, NULL );",
	"SELECT t.* FROM odbxtest AS t;",
	"SELECT * FROM odbxtest WHERE string = 'empty';",
	"SELECT * FROM odbxtest WHERE i16 = '2';",
	"DELETE FROM odbxtest WHERE string LIKE '%string';",
	"DROP TABLE odbxtest;",
	NULL
};

static const char* mysql_multi[] = {
	"CREATE TABLE odbxtest ( i8 tinyint, i16 smallint, i32 int, i64 bigint, f4 float, f8 double, string varchar(25) ) ENGINE InnoDB;",
	"INSERT INTO odbxtest VALUES ( 100, 1000, 100000, 10000000, 10000.1, 10000000.1, 'first string' );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, 'varstring' );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, NULL );",
	"SELECT t.* FROM odbxtest AS t; SELECT * FROM odbxtest WHERE string = 'empty';",
	"START TRANSACTION; DELETE FROM odbxtest WHERE i16 = '2'; ROLLBACK;",
	"SELECT * FROM odbxtest WHERE i16 = '2';",
	"DELETE FROM odbxtest WHERE string LIKE '%string';",
	"DROP TABLE odbxtest;",
	NULL
};

static const char** mysql_stmt[] = { mysql_basic, mysql_multi };


static const char* pgsql_multi[] = {
	"CREATE TABLE odbxtest ( int16 smallint, int32 integer, int64 bigint, float4 real, float8 double precision, string varchar(25) );",
	"INSERT INTO odbxtest VALUES ( 1000, 100000, 10000000, 10000.1, 10000000.1, 'first string' );",
	"INSERT INTO odbxtest VALUES ( 2, 4, 8, 4.0, 8.0, 'varstring' );",
	"INSERT INTO odbxtest VALUES ( 2, 4, 8, 4.0, 8.0, NULL );",
	"SELECT t.* FROM odbxtest AS t;",
	"SELECT * FROM odbxtest WHERE string = 'empty';",
	"BEGIN; DELETE FROM odbxtest WHERE int16 = 2; ABORT;",
	"SELECT * FROM odbxtest WHERE int16 = '2';",
	"DELETE FROM odbxtest WHERE string LIKE '%string';",
	"DROP TABLE odbxtest;",
	NULL
};

static const char** pgsql_stmt[] = { pgsql_multi, pgsql_multi };


static const char* sqlite_multi[] = {
	"CREATE TABLE odbxtest ( int8 tinyint, int16 smallint, int32 integer, int64 bigint, float4 float, float8 double, string varchar(25) );",
	"INSERT INTO odbxtest VALUES ( 100, 1000, 100000, 10000000, 10000.1, 10000000.1, 'first string' );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, 'varstring' );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, NULL );",
	"INSERT INTO odbxtest VALUES ( 1, 2, 4, 8, 4.0, 8.0, 'first string' );",
	"SELECT ROWID, t.* FROM odbxtest AS t;",
	"SELECT * FROM odbxtest WHERE string LIKE 'first string';",
	"SELECT * FROM odbxtest WHERE string = 'empty';",
	"BEGIN TRANSACTION; DELETE FROM odbxtest WHERE int16 = 2; ROLLBACK TRANSACTION;",
	"SELECT * FROM odbxtest WHERE int16 = '2';",
	"DELETE FROM odbxtest WHERE string LIKE '%string';",
	"DROP TABLE odbxtest;",
	NULL
};

static const char** sqlite_stmt[] = { sqlite_multi, sqlite_multi };



struct odbxtest
{
	char* name;
	int value;
	int result;
	int try;
};


static struct odbxtest cap[] = {
	{ "ODBX_CAP_BASIC", ODBX_CAP_BASIC, 0, 1 },
};

#define CAPMAX 1


static struct odbxtest opt[] = {
	{ "ODBX_OPT_API_VERSION", ODBX_OPT_API_VERSION, 0, 0 },
	{ "ODBX_OPT_THREAD_SAFE", ODBX_OPT_THREAD_SAFE, 0, 0 },
	{ "ODBX_OPT_TLS", ODBX_OPT_TLS, 0, 0 },
	{ "ODBX_OPT_MULTI_STATEMENTS", ODBX_OPT_MULTI_STATEMENTS, 0, 1 },
	{ "ODBX_OPT_PAGED_RESULTS", ODBX_OPT_PAGED_RESULTS, 0, 1 },
	{ "ODBX_OPT_COMPRESS", ODBX_OPT_COMPRESS, 0, 1 }
};

#define OPTMAX 6





int exec( odbx_t* handle, char* qptr[], int verbose )
{
	int i, err = 0;
	struct timeval tv;
	odbx_result_t* result;
	unsigned long fields;


	while( *qptr != NULL )
	{
		if( verbose ) { fprintf( stdout, "  odbx_query(): '%s'\n", *qptr ); }
		if( ( err = odbx_query( handle, *qptr, strlen( *qptr ) ) ) < 0 )
		{
			fprintf( stderr, "odbx_query(): %s\n", odbx_error( handle, err ) );
			return err;
		}

		tv.tv_sec = 3;
		tv.tv_usec = 0;

		/* use values >1 for paged output (if supported) or 0 for all rows */
		while( ( err = odbx_result( handle, &result, &tv, 0 ) ) != 0 )
		{
			if( verbose ) { fprintf( stdout, "  odbx_result()\n" ); }

			if( err < 0 )
			{
				fprintf( stderr, "    odbx_result(): %s\n", odbx_error( handle, err ) );
				return err;
			}

			if( err == 1 )
			{
				fprintf( stdout, "    odbx_result(): Timeout\n" );
				continue;
			}

			if( err == 2 )
			{
				fprintf( stdout, "    Affected rows: %d\n", odbx_rows_affected( result ) );
				odbx_result_free( result );
				continue;
			}

			if( verbose ) { fprintf( stdout, "  odbx_column_count()\n" ); }
			fields = odbx_column_count( result );

			while( ( err = odbx_row_fetch( result ) ) != 0 )
			{
				if( verbose ) { fprintf( stdout, "  odbx_row_fetch()\n" ); }

				if( err < 0 )
				{
					fprintf( stderr, "odbx_row_fetch(): %s\n", odbx_error( handle, err ) );
					break;
				}

				for( i = 0; i < fields; i++ )
				{
					if( ( err = odbx_column_type( result, i ) ) < 0 )
					{
						fprintf( stderr, "odbx_column_type(): %s\n", odbx_error( handle, err ) );
						continue;
					}

					fprintf( stdout, "    column %d: %s = %s\n", i, odbx_column_name( result, i ), odbx_field_value( result, i ) );
				}
			}

			if( verbose ) { fprintf( stdout, "  odbx_result_free()\n" ); }
			odbx_result_free( result );

			tv.tv_sec = 3;
			tv.tv_usec = 0;
		}

		qptr++;
	}

	return 0;
}



void help( char* name )
{
	fprintf( stdout, "Usage: %s <options>\n", name );
	fprintf( stdout, "   -b <backend>    Database backend (mysql, pgsql, sqlite, sqlite3)\n" );
	fprintf( stdout, "   -h <server>     Server name, IP address or directory\n" );
	fprintf( stdout, "   -p <port>       Server port\n" );
	fprintf( stdout, "   -d <database>   Database or file name\n" );
	fprintf( stdout, "   -u <user>       User name for authentication\n" );
	fprintf( stdout, "   -w <password>   Password for authentication\n" );
	fprintf( stdout, "   -e              Force encrypted connection\n" );
	fprintf( stdout, "   -r              Number of runs\n" );
	fprintf( stdout, "   -v              Verbose mode\n" );
}



int main( int argc, char* argv[] )
{
	int i, run = 0;
	int option, err = 0;
	odbx_t* handle = NULL;
	char*** stmts = NULL;
	char** queries = NULL;
	char *backend, *host, *port, *db, *user, *pass;
	int param, verbose = 0, encrypt = 0, runs = 1;


	backend = host = port = db = user = pass = NULL;

	while( ( param = getopt( argc, argv, "b:h:p:d:u:w:er:v" ) ) != -1 )
	{
		switch( param )
		{
			case 'b':
				backend = strdup( optarg );
				break;
			case 'h':
				host = strdup( optarg );
				break;
			case 'p':
				port = strdup( optarg );
				break;
			case 'd':
				db = strdup( optarg );
				break;
			case 'u':
				user = strdup( optarg );
				break;
			case 'w':
				pass = strdup( optarg );
				break;
			case 'e':
				encrypt = 1;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'r':
				if( optarg != NULL ) { runs = (int) strtol( optarg, NULL, 10 ); }
				break;
			default:
				fprintf( stderr, "Unknown option '%c' with arg '%s'\n", (char) param, optarg );
				return 1;
		}
	}

	if( backend == NULL ) { help( argv[0] ); err = 1; goto ERROR; }

	if( !strncmp( backend, "mysql", 5 ) ) { stmts = (char***) mysql_stmt; }
	if( !strncmp( backend, "pgsql", 5 ) ) { stmts = (char***) pgsql_stmt; }
	if( !strncmp( backend, "sqlite", 6 ) ) { stmts = (char***) sqlite_stmt; }

	fprintf( stdout, "\nParameter:\n" );
	fprintf( stdout, "  backend=%s\n", backend );
	fprintf( stdout, "  host=%s\n", host );
	fprintf( stdout, "  port=%s\n", port );
	fprintf( stdout, "  database=%s\n", db );
	fprintf( stdout, "  user=%s\n", user );
	fprintf( stdout, "  password=%s\n", pass );
	fprintf( stdout, "  encrypt=%d\n", encrypt );
	fprintf( stdout, "  runs=%d\n", runs );
	fprintf( stdout, "  verbose=%d\n",verbose );


	for( run = 0; run < 1; run++ )
	{
		if( verbose ) { fprintf( stdout, "\n%d. Run:\n", run+1 ); }

		if( verbose ) { fprintf( stdout, "  odbx_init()\n" ); }
 		if( ( err = odbx_init( &handle, backend, host, port ) ) < 0 )
		{
			fprintf( stderr, "odbx_init(): %s\n", odbx_error( handle, err ) );
			goto FAIL;
		}

		if( verbose ) { fprintf( stdout, "  odbx_capabilities()\n" ); }
		for( i = 0; i < CAPMAX; i++ )
		{
			if( ( err = odbx_capabilities( handle, (unsigned int) cap[i].value ) ) < 0 )
			{
				fprintf( stderr, "odbx_capabilities(): %s\n", odbx_error( handle, err ) );
				goto ERROR;
			}

			cap[i].result = err;
			fprintf( stdout, "    %s: %d\n", cap[i].name, cap[i].result );
		}

		if( verbose ) { fprintf( stdout, "  odbx_{get/set}_option()\n" ); }
		for( i = 0; i < OPTMAX; i++ )
		{
			if( ( err = odbx_get_option( handle, opt[i].value, (void*) &(opt[i].result) ) ) < 0 )
			{
				fprintf( stderr, "odbx_get_option(): %s\n", odbx_error( handle, err ) );
				goto ERROR;
			}

			fprintf( stdout, "    %s: %d %s\n", opt[i].name, opt[i].result, opt[i].result && opt[i].try ? "(using)" : "" );

			if( opt[i].result && opt[i].try && ( err = odbx_set_option( handle, opt[i].value, (void*) &(opt[i].try) ) ) < 0 )
			{
				fprintf( stderr, "odbx_set_option(): %s\n", odbx_error( handle, err ) );
				goto ERROR;
			}
		}

		option = ODBX_TLS_ALWAYS;
		if( encrypt && ( err = odbx_set_option( handle, ODBX_OPT_TLS, (void*) &option ) ) < 0 )
		{
			fprintf( stderr, "odbx_set_option(): %s\n", odbx_error( handle, err ) );
			goto ERROR;
		}

		queries = stmts[0];
		if( ( err = odbx_get_option( handle, ODBX_OPT_MULTI_STATEMENTS, (void*) &option ) ) < 0 )
		{
			fprintf( stderr, "odbx_get_option(): %s\n", odbx_error( handle, err ) );
			goto ERROR;
		}
		queries = stmts[option];

		if( verbose ) { fprintf( stdout, "  odbx_bind_simple()\n" ); }
		if( ( err = odbx_bind_simple( handle, db, user, pass ) ) < 0 )
		{
			fprintf( stderr, "odbx_bind_simple(): %s\n", odbx_error( handle, err ) );
			goto ERROR;
		}

		for( i = 0; i < runs; i++ )
		{
			if( ( err = exec( handle, queries, verbose ) ) < 0 )
			{
				fprintf( stdout, "exec() failed in run %d\n", i+1 );
				break;
			}
		}

		if( verbose ) { fprintf( stdout, "  odbx_unbind()\n" ); }
		if( ( err = odbx_unbind( handle ) ) < 0 )
		{
				fprintf( stderr, "odbx_unbind(): %s\n", odbx_error( handle, err ) );
		}

ERROR:

		if( verbose ) { fprintf( stdout, "  odbx_finish()\n" ); }
		if( ( err = odbx_finish( handle ) ) < 0 )
		{
				fprintf( stderr, "odbx_finish(): %s\n", odbx_error( handle, err ) );
		}
	}

FAIL:

	if( backend != NULL ) { free( backend ); }
	if( host != NULL ) { free( host ); }
	if( port != NULL ) { free( port ); }
	if( db != NULL ) { free( db ); }
	if( user != NULL ) { free( user ); }
	if( pass != NULL ) { free( pass ); }

	return err;
}
